import json

from remotemanager.utils import ensure_filetype
from remotemanager.serialisation.serial import serial


class serialjson(serial):
    """
    subclass of serial, implementing json methods
    """

    def dump(self, obj, file):
        file = ensure_filetype(file, self.extension)
        with open(file, 'w+') as ofile:
            json.dump(obj, ofile)

    def load(self, file):
        file = ensure_filetype(file, self.extension)
        with open(file, 'r') as ofile:
            data = json.load(ofile)
        return data

    @property
    def extension(self):
        return ".json"

    @property
    def importstring(self):
        return "import json"

    @property
    def callstring(self):
        return "json"



import copy
import os
import typing
import weakref

from remotemanager.storage.database import Database
from remotemanager.storage.function import Function
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.logging.utils import format_iterable
from remotemanager.utils.uuid import generate_uuid
from remotemanager.utils import ensure_list
from remotemanager.logging import LoggingMixin

from datetime import datetime


class Runner(SendableMixin, LoggingMixin):
    """
    The Runner class stores any info pertaining to this specific run. E.g.
    Arguments, result, run status, files, etc.

    .. warning::
        Interacting with this object directly could cause unstable behaviour.
        It is best to allow Dataset to handle the runners. If you require a
        single run, you should create a Dataset and append just that one run.
    """

    _defaults = {'skip': False}

    _state_direct_remote = 'disconnected - remote direct run'
    _state_script_remote = 'script submitted remotely'
    _state_script_local = 'script submitted locally'

    def __init__(self,
                 arguments: dict,
                 dbfile: str,
                 parent_uuid: str,
                 extra_files_send: list = None,
                 extra_files_recv: list = None,
                 dependency_key: str = None,
                 **kwargs):

        self._run_options = self._set_defaults(kwargs)

        self._args = arguments

        self._extra_files = {'send': extra_files_send
                                if extra_files_send is not None else [],
                             'recv': extra_files_recv
                                if extra_files_recv is not None else []}

        if arguments is not None \
                and not isinstance(arguments, dict):
                raise ValueError(f'runner arguments ({type(arguments)}) '
                                 'must be dict-type')

        self._dependency_info = {}

        uuid_slug = copy.deepcopy(arguments)
        if uuid_slug is None:
            uuid_slug = {}
        uuid_slug.update(kwargs)
        uuid_slug.update(self.extra_files)
        uuid_slug.update({'dependency_key': dependency_key})
        self._uuid = generate_uuid(format_iterable(uuid_slug))
        self._time_format = '%Y-%m-%d %H:%M:%S'
        self._state_time = None
        self._state = None
        self._extension = 'yaml'

        self._dbfile = dbfile
        self._parent_uuid = parent_uuid

        self._logger.info(f'new runner (id {self.uuid}) created')

        self._history = {}
        self.state = 'created'

    def __hash__(self) -> str:
        return hash(self.uuid)

    @property
    def database(self) -> Database:
        """
        Access to the stored database object.
        Creates a connection if none exist.

        Returns:
            Database
        """
        if not hasattr(self, '_database'):
            self._database = Database(file=self._dbfile)
        return self._database

    @staticmethod
    def _set_defaults(kwargs: dict = None) -> dict:
        """
        Sets default arguments as expected. If used as a staticmethod, returns
        the defaults
        """

        if kwargs is None:
            kwargs = {}

        for k, v in Runner._defaults.items():
            if k not in kwargs:
                kwargs[k] = v

        return kwargs

    @property
    def uuid(self):
        """
        The uuid of this runner
        """
        return self._uuid

    @property
    def short_uuid(self):
        """
        A short uuid for filenames
        """
        return self.uuid[:8]

    @property
    def runfile(self):
        """
        Filename of the python runfile
        """
        return f'{self.short_uuid}-run.py'

    def runpath(self, local: bool = False):
        """
        Runfile name

        Args:
            local (bool):
                returns the runfile path prior to sending if True
        """
        if local:
            return os.path.join(self.local_dir, self.runfile)
        paths = [self.remote_dir]
        if self.run_dir != self.remote_dir and self.run_dir is not None:
            paths.append(self.run_dir)
        paths.append(self.runfile)
        return os.path.join(*paths)

    @property
    def scriptfile(self):
        """
        Filename of the run script
        """
        return f'{self.short_uuid}-run.sh'

    def scriptpath(self, local: bool = False):
        """
        Run script path

        Args:
            local (bool):
                returns the run script path prior to sending if True
        """
        if local:
            return os.path.join(self.local_dir, self.scriptfile)
        paths = [self.remote_dir]
        paths.append(self.scriptfile)
        return os.path.join(*paths)

    @property
    def resultfile(self):
        """
        Result file name
        """
        return f'{self.short_uuid}-result.{self.result_extension}'

    def resultpath(self, local: bool = False):
        """
        Path to result file
        """
        if local:
            return os.path.join(self.local_dir, self.resultfile)
        paths = [self.remote_dir]
        if self.run_dir != self.remote_dir and self.run_dir is not None:
            paths.append(self.run_dir)
        paths.append(self.resultfile)
        return os.path.join(*paths)

    @property
    def result_extension(self):
        """
        Resultfile file format extension
        """
        return self._extension

    @result_extension.setter
    def result_extension(self, ext):
        """
        Sets the resultfile format extension

        .. warning::
            This does not change anything aside from the extension that
            the runner looks for when trying to find a result file. If you
            require a different serialisation, you should set the serialiser.
        """
        self._extension = ext.strip('.')

    @property
    def local_dir(self):
        """
        Local staging directory
        """
        return self._run_options.get('local_dir', 'staging')

    @local_dir.setter
    def local_dir(self,
                  path: str) -> None:
        """
        Sets the local_dir
        """
        self._run_options['local_dir'] = path

    @property
    def remote_dir(self):
        """
        Target directory on the remote for transports
        """
        if 'remote_dir' in self._run_options:
            return self._run_options['remote_dir']
        return self._run_options.get('run_dir', 'runner_remote')

    @remote_dir.setter
    def remote_dir(self,
                   path: str) -> None:
        """
        Sets the remote_dir
        """
        self._logger.debug(f'setting remote dir to {path}')
        self._run_options['remote_dir'] = path

    @property
    def run_dir(self):
        """
        Intended running directory. If not set, uses remote_dir

        .. note::
            If both remote_dir and run_dir are set, the files will be
            transferred to remote_dir, and then executed within run_dir
        """
        if 'run_dir' in self._run_options:
            return self._run_options['run_dir']
        return None

    @run_dir.setter
    def run_dir(self,
                dir: str) -> None:
        """
        Sets the run_dir
        """
        self._run_options['run_dir'] = dir

    @property
    def args(self):
        """
        Arguments for the function
        """
        return self._args

    @property
    def extra_files(self):
        """
        Returns the extra files set for this runner
        """
        return self._extra_files

    @property
    def result(self):
        """
        Result (If available)
        """
        if hasattr(self, '_result'):
            return self._result
        return None

    @result.setter
    def result(self, result) -> None:
        """
        Creates and sets the result property, setting the state to "completed"

        Args:
            result:
                run result
        """
        self._result = result
        self.state = 'completed'

    def clear_result(self):
        """
        Removes any results, and sets the state to "result wiped"
        """
        try:
            del self._result
        except AttributeError:
            pass

        def remove_file(path):
            self._logger.info(f'attempting to clear result file {path}')
            try:
                os.remove(path)
                self._logger.info('Done')
            except FileNotFoundError:
                self._logger.info('file not found')

        remove_file(self.resultpath(local=True))
        remove_file(self.resultpath())

        self.state = 'result wiped'

    @property
    def state(self):
        """
        Returns the most recent runner state
        """
        return self._state

    @state.setter
    def state(self,
              newstate: str) -> None:
        """
        Update the state and store within the runner history
        """

        state_time = datetime.now()

        self.insert_history(state_time, newstate)

        self._state_time = int(state_time.strftime('%s'))
        self._state = newstate

    def format_time(self, time: datetime.time) -> str:
        """
        Format the datetime object into a dict key

        Args:
            time (datetime.time):
                time object to be formatted to string

        Returns:
            (str):
                formatted time
        """
        return time.strftime(self._time_format)

    @property
    def history(self) -> dict:
        """
        State history of this runner
        """
        return self._history

    @property
    def status_list(self) -> list:
        """
        Returns a list of status updates
        """
        return list(self._history.values())

    def insert_history(self,
                       time: datetime,
                       newstate: str) -> None:
        """
        Insert a state into this runner's history

        Args:
            time (datetime.time):
                time this state change occurred
            newstate (str):
                status to update
        """
        if not isinstance(time, datetime):
            raise ValueError(f'time of type {type(time)} should be a datetime '
                             f'instance')

        base_timekey = self.format_time(time)
        idx = 0
        timekey = f'{base_timekey}/{idx}'
        while timekey in self._history:
            idx += 1

            timekey = f'{base_timekey}/{idx}'
            self._logger.info(f'timekey updated to {timekey}')

        self._logger.info(f'updating runner {self.short_uuid} state -> '
                          f'{newstate}')
        self._history[timekey] = newstate

    @property
    def last_updated(self):
        """
        Time that this runner state last changed
        """
        return self._state_time

    def run(self,
            dry_run: bool = False,
            **kwargs):
        """
        Perform a manual run

        .. warning::
            This method should be used sparingly, as it creates a Datset
            object within the function from the Database. This is a costly
            process and potentially unstable.

        Args:
            dry_run (bool):
                create files, but do not run
        """

        parent = self.unserialise(self.database.find(self._parent_uuid))

        self._run_options.update(kwargs)
        write_success = self._write_runfile(parent.function,
                                            parent.serialiser,
                                            **self._run_options)
        if not write_success:
            return

        parent.transport.queue_for_push(self.runfile,
                                        self.local_dir,
                                        self.remote_dir)

        script = parent._script_sub(**self.run_args)

        self._write_script(parent.url.python,
                           script)

        parent.transport.queue_for_push(self.scriptfile,
                                        self.local_dir,
                                        self.remote_dir)

        cmd = f'cd {self.remote_dir} && {parent.submitter} {self.scriptfile}'

        if not dry_run:
            parent.transport.transfer()
            parent.url.cmd(cmd, asynchronous=False)
            self.state = 'submitted'
        else:
            parent.transport.wipe_transfers()
            print(cmd)

    def _assess_run(self):
        self._logger.info(f'assessing run for runner {self.short_uuid}')
        self._logger.info('run args:')
        self._logger.info(format_iterable(self.run_args))

        if self.state == 'submitted' and not self.run_option('force', False):
            self._logger.info('skipping already submitted run')
            print('already submitted, skipping')
            return False

        # old skip test unpacks a runner from the database
        if self.is_finished and not self.run_option('skip', True):
            self._logger.info('skipping already completed run')
            print('run already completed, skipping')
            return False

        self._logger.info('allowing run')
        return True

    def _write_runfile(self,
                       function,
                       serialiser,
                       write_file: bool = True,
                       **kwargs) -> str:
        """
        Writes the python file which actually runs the function

        Args:
            function (Function):
                stored function object
            serialiser (Serialiser):
                stored serialiser object
            **kwargs:
                run arguments
        """
        self._run_options.update(kwargs)
        allow = self._assess_run()

        if not allow:
            return None

        self._logger.info(f'pre-running function {self.uuid}')

        self.update_run_options(kwargs)

        runscript = [serialiser.importstring,
                    function.dump_to_string(self.args)] + \
                    ['\t' + line for line in
                    serialiser.dumpstring(self.resultfile)]

        if self._dependency_info.get('child', False):
            runscript.insert(1, self._dependency_info["parent_import"])
        if self._dependency_info.get('parent', False):
            runscript.append(self._dependency_info['child_submit'] + '\n')

        output = '\n'.join(runscript)
        if write_file:
            if not os.path.isdir(self.local_dir):
                self._logger.info(f'creating local dir {self.local_dir}')
                os.makedirs(self.local_dir)
            with open(self.runpath(local=True), 'w+') as o:
                o.write(output)

        return output

    def _write_script(self,
                      python: str,
                      script: str,
                      write_file: bool = True) -> str:
        """
        Writes the jobscript for this runner

        Args:
            python (str):
                python command to launch runfile
            script (str):
                script header
        """
        tmp = []
        if isinstance(script, str):
            tmp.append(script)
        else:
            tmp.append(script(**self.run_args))

        if self.run_dir and self.run_dir != self.remote_dir:
            cmd = f'mkdir -p {self.run_dir} && ' \
                  f'cd {self.run_dir} && ' \
                  f'{python} ../{self.runfile}'
        else:
            cmd = f'{python} {self.runfile}'  # run file

        tmp.append(cmd)

        path = self.scriptpath(local=True)
        output = '\n'.join(tmp)
        if write_file:
            self._logger.info(f'writing run script to {path}')
            with open(path, 'w+') as o:
                o.write(output)

        return output

    @property
    def is_finished(self):
        """
        Attempts to determine if this runner has completed its run

        Returns (bool):
            completion status
        """
        self._logger.info(f'checking finished state of runner '
                          f'{self.short_uuid}')
        return hasattr(self, '_result')

    def update_run_options(self,
                           run_args: dict) -> None:
        """
        Update run args with dict `run_args`
        Args:
            run_args (dict):
                new run arguments

        Returns:
            None
        """
        self._logger.info('updating run options with new run args:')
        self._logger.info(format_iterable(run_args))

        self._run_options.update(run_args)

    def run_option(self,
                   option: str,
                   default=None):
        """
        Return a run option

        Args:
            option (str):
                key to search for
            default:
                default argument to provide to get

        Returns:
            option if available, else None
        """
        ret = self._run_options.get(option, default)
        self._logger.debug(f'getting run option {option}: {ret}')
        return ret

    @property
    def run_args(self):
        """
        Display the run arguments

        Returns:
            (dict) run_args
        """
        return self._run_options



import logging
import re

from remotemanager.storage.sendablemixin import SendableMixin


class Flags(SendableMixin):
    """
    Basic but flexible handler for terminal command flags

    Allows for inplace modification:

    >>> f = Flags('abcdd')
    >>> f -= 'd'
    >>> f.flags
    >>> '-abcd'

    Arguments:
        initial_flags (str):
            initial base flags to be used and modified if needed
        prefix (optional, str):
            the non-argument prefix for these flags (`-`, `--`, etc.)
    """

    _logger = logging.getLogger(__name__ + '.Flags')

    def __init__(self, initial_flags: str = '', prefix: str = None):

        self._logger.debug(f'creating Flags with initial flags '
                           f'{initial_flags} and prefix {prefix}')

        self._prefix = ''
        if prefix is None:
            if '-' in initial_flags:
                self.prefix = '-' * initial_flags.count('-')
            else:
                self.prefix = '-'
            self._logger.debug(f'prefix set to {self.prefix}')
        else:
            self.prefix = prefix
            initial_flags = strip_non_alphanumeric(initial_flags)

        self.flags = initial_flags

    def __repr__(self):
        return f'Flags({self.flags})'

    def __add__(self, other):
        self._flags = self._flags + other.strip('-')
        self._logger.debug(f'adding {other} to flags. '
                           f'Flags are now {self.flags}')

    def __iadd__(self, other):
        self.__add__(other)
        self._logger.debug(f'adding {other} to flags inplace. '
                           f'Flags are now {self.flags}')
        return self

    def __sub__(self, other):
        """Subtract unique flags in `other` once."""
        for char in ''.join(set(other)):
            self._flags = self._flags.replace(char, '', 1)
        self._logger.debug(f'subtracting {other} from flags. '
                           f'Flags are now {self.flags}')

    def __isub__(self, other):
        self.__sub__(other)
        self._logger.debug(f'subtracting {other} from flags inplace. '
                           f'Flags are now {self.flags}')
        return self

    @property
    def flags(self):
        """Returns the fully qualified flags as a string"""
        if len(self._flags) == 0:
            return ''
        return self.prefix + self._flags

    @flags.setter
    def flags(self, inp):
        """Set the flags to the new input

        Arguments:
            inp (str):
                new flags to use (will overwrite old ones)
        """
        self._logger.debug(f'setting flags to {inp}')
        for char in set(self.prefix):
            inp = inp.strip(char)
        self._logger.debug(f'after cleaning, flags are now: {inp}')
        self._flags = inp

    @property
    def prefix(self):
        """
        Attribute determining the - prefix used for these flags

        For example, a flag with a prefix of `-` will be added as `-flag`
        Change the prefix to `--` to be passed as `--flag`, and so on
        """
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        self._prefix = prefix


def strip_non_alphanumeric(string):
    """
    remove any non-alphanumeric strings from input string

    Args:
        string (str):
            input string

    Returns (str):
        input string, sans any non-alphanumeric chars

    """
    pattern = re.compile('[\W_]+', re.UNICODE)

    return pattern.sub('', string)



import yaml

from remotemanager.utils import ensure_filetype
from remotemanager.serialisation.serial import serial


class serialyaml(serial):
    """
    subclass of serial, implementing yaml methods
    """

    def dump(self, obj, file):
        file = ensure_filetype(file, self.extension)
        with open(file, 'w+') as ofile:
            yaml.dump(obj, ofile)

    def load(self, file):
        file = ensure_filetype(file, self.extension)
        with open(file, 'r') as ofile:
            data = yaml.safe_load(ofile)
        return data

    @property
    def extension(self):
        return ".yaml"

    @property
    def importstring(self):
        return "import yaml"

    @property
    def callstring(self):
        return "yaml"

    def dumpstring(self, file: str) -> list:
        """
        Python code for dumping an object named `result`

        Args:
            file (str):
                filename to dump to

        Returns (list):
            Formatted list of strings for run files.
            See subclasses for examples
        """
        file = ensure_filetype(file, self.extension)
        string = ["if isinstance(result, tuple):",
                  "\tresult = list(result)",
                  f"with open('{file}', 'w+') as o:",
                  f'\t{self.callstring}.dump(result, o)']
        return string

    def loadstring(self, file: str) -> list:
        """
        Python code for loading an object from file `file` into object named
        `loaded`

        Args:
            file (str):
                filename to load from

        Returns (list):
            Formatted list of strings for run files.
            See subclasses for examples
        """
        file = ensure_filetype(file, self.extension)
        string = [f"with open('{file}', 'r') as o:",
                  f'\tloaded = {self.callstring}.safe_load(o)']
        return string



from remotemanager.utils import ensure_filetype
from remotemanager.storage.sendablemixin import SendableMixin


class serial(SendableMixin):

    """
    Baseclass for holding serialisation methods. Subclass this class when
    implementing new serialisation methods
    """

    def __init__(self):
        pass

    def dump(self, obj, file: str) -> None:
        """
        Dumps object `obj` to file `file`

        Args:
            obj:
                object to be dumped
            file (str):
                filepath to dump to

        Returns:
            None
        """
        raise NotImplementedError

    def load(self, file: str):
        """
        Loads previously dumped data from file `file`

        Args:
            file (str):
                filepath to load

        Returns:
            Stored object
        """
        raise NotImplementedError

    @property
    def extension(self) -> str:
        """
        Returns (str):
            intended file extension
        """
        raise NotImplementedError

    @property
    def importstring(self) -> str:
        """
        Returns (str):
            Module name to import.
            See subclasses for examples
        """
        raise NotImplementedError

    @property
    def callstring(self) -> str:
        """
        Returns (str):
            Intended string for calling this module's dump.
            See subclasses for examples
        """
        raise NotImplementedError

    def dumpstring(self, file: str) -> list:
        """
        Python code for dumping an object named `result`

        Args:
            file (str):
                filename to dump to

        Returns (list):
            Formatted list of strings for run files.
            See subclasses for examples
        """
        file = ensure_filetype(file, self.extension)
        string = [f"with open('{file}', 'w+') as o:",
                  f'\t{self.callstring}.dump(result, o)']
        return string

    def loadstring(self, file: str) -> list:
        """
        Python code for loading an object from file `file` into object named
        `loaded`

        Args:
            file (str):
                filename to load from

        Returns (list):
            Formatted list of strings for run files.
            See subclasses for examples
        """
        file = ensure_filetype(file, self.extension)
        string = [f"with open('{file}', 'r') as o:",
                  f'\tloaded = {self.callstring}.load(o)']
        return string




from remotemanager.connection.url import URL
from remotemanager.utils import safe_divide


class Summer(URL):
    """
    subclassed URL specialising in connecting to the CEA Summer HPC
    """

    def __init__(self, **kwargs):

        if kwargs.get('host', None) is not None:
            raise ValueError('cannot change host of dedicated URL')

        kwargs['host'] = 'summer.intra.cea.fr'
        # count of np attribute of pbsnodes
        self._nodes = {8: 60,
                       12: 16,
                       16: 76,
                       32: 52,
                       40: 44}

        super().__init__(**kwargs)

        self.submitter = 'qsub'
        self.pragma = '#PBS'

        self.sourcepath = '/W/$USER/build/bigdft/bigdftvars.sh'

    @property
    def ssh(self):
        """
        Returns (str):
            modified ssh string for Summer avoiding perl error
        """
        return 'LANG=C ' + super().ssh

    def script(self,
               mpi: int,
               omp: int,
               nodes: int,
               name: str,
               queue: str,
               walltime: str,
               shebang: str = '#!/bin/bash',
               outfile: str = None,
               errfile: str = None,
               sourcepath: str = None,
               extra_lines: str = None,
               **kwargs) -> str:
        """
        Takes job arguments and produces a valid Summer jobscript

        Args:
            mpi (int):
                mpi for this job (NCPUs per machine)
            omp (int):
                omp for this job (Threads per task)
            nodes (int):
                number of nodes for this job (NMachines)
            name (str):
                job name for scheduler
            queue (str):
                queue for this job
            walltime (str):
                walltime in HH:MM:SS format
            shebang (str):
                script shebang
                defaults to 'bin/bash'
            outfile (str):
                file to place any scheduler stdout
                defaults to {name}-stdout
            errfile (str):
                file to place any scheduler stderr
                defaults to {name}-stderr
            sourcepath (str):
                path to executable
            extra_lines (str):
                any extra script lines to add
            kwargs:
                required to catch any extra arguments which may not be meant
                for the script

        Returns:
            (str):
                script
        """
        script = [shebang]

        outfile = outfile or f'{name}-stdout'
        errfile = errfile or f'{name}-stderr'
        sourcepath = sourcepath or self.sourcepath

        options = {'-N': name,
                   '-q': queue,
                   '-o': outfile,
                   '-e': errfile,
                   '-l': f'nodes={nodes}:'
                         f'ppn={mpi},'
                         f'walltime={walltime}'}

        modules = ['icc',
                   'impi',
                   'mkl/18',
                   'python/anaconda3']

        for flag, value in options.items():
            script.append(f'{self.pragma} {flag} {value}')

        script.append('')
        for module in modules:
            script.append(f'module load {module}')

        postscript = f"""
cd $PBS_O_WORKDIR
export OMP_NUM_THREADS={omp}

source {sourcepath}
export BIGDFT_MPIRUN='mpirun'
export FUTILE_PROFILING_DEPTH=0
"""

        script.append(postscript)
        if extra_lines is not None:
            script.append(extra_lines)

        return '\n'.join(script)



import collections.abc
import importlib
import logging
import os.path

from remotemanager.logging.utils import format_iterable

import yaml

CLASS_STORAGE_KEY = '~serialisedclass~'
TUPLE_STORAGE_KEY = '~serialisedtuple~'
SET_STORAGE_KEY = '~serialisedset~'

_logger = logging.getLogger(__name__)


class SendableMixin:
    """
    baseclass for a "sendable" object, providing methods for conversion to
    yaml format
    """

    def pack(self,
             uuid: str = None,
             file: str = None):
        """
        "packs" the object into a dict-format, ready for yaml-dump

        Args:
            uuid (str):
                optional uuid string to package this data inside
            file (str):
                package to file

        Returns:
            (dict):
                packaged Runner
        """
        # remove any attributes not to be packaged
        # these objects are either not needed, or can be re-created on the
        # recipient end
        never_package = ['_logger', '_logobj']

        do_not_package = {'Dataset': ['_database'],
                          'URL': ['_urlutils'],
                          'CMD': ['_subprocess'],
                          'Dependency': ['_parent', '_child']}

        class_storage = get_class_storage(self)

        # if the object is a subclass, find the name of the parent by getting
        # the mro of the object, and extracting it from the set
        parent_class_names = get_mro_classnames(self)
        sub_block = set(do_not_package) & set(parent_class_names)

        name_to_block = None
        if class_storage['name'] in do_not_package:
            # classname has extra attrs to block
            name_to_block = class_storage['name']
        elif len(sub_block) != 0:
            # parent class has extra attrs to block
            name_to_block = list(sub_block)[0]

        if name_to_block:
            block = never_package + do_not_package[name_to_block]
        else:
            block = never_package

        payload = {self.serialise(k): self.serialise(v)
                   for k, v in self.__dict__.items()
                   if k not in block}

        payload[CLASS_STORAGE_KEY] = class_storage

        if uuid:

            if '_uuid' in payload and payload['_uuid'] != uuid:
                raise ValueError('passed uuid and _uuid key in payload do'
                                 'not match!')

            payload['_uuid'] = uuid

            payload = {uuid: payload}

        if file is not None:
            print(f'dumping payload to {file}')
            with open(file, 'w+') as o:
                yaml.dump(payload, o)

            return None
        return payload

    @classmethod
    def unpack(cls,
               data: dict = None,
               file: str = None,
               limit: bool = True):
        """
        Re-create an object from a packaged payload coming from obj.pack

        .. note ::

            use this function to unpack from a payload _outside_ an object

        Args:
            data (dict):
                __dict__ payload from the object that was packaged
            file (str):
                filepath to unpack from, if data is not given
            limit (bool):
                disable to allow outside classes to be unserialised

        Returns:
            re-created object
        """

        if data is None:
            if file is None:
                raise ValueError('please provide a data payload or filepath '
                                 'to read from')
            elif not os.path.isfile(file):
                raise FileNotFoundError(f'could not find file {file}')

            with open(file, 'r') as o:
                data = yaml.safe_load(o)

        # if the data passed is a {uuid: {data}} format dict, search for a uuid
        # within and extract the {data}
        try:
            firstkey = list(data)[0]
            if isinstance(data[firstkey], dict) \
                    and '_uuid' in data[firstkey] \
                    and data[firstkey]['_uuid'] == firstkey:
                data = data[firstkey]
        except IndexError:
            # if there's no keys in data, then we definitely don't have a uuid
            pass

        # create a new instance of this class
        new = cls.__new__(cls)

        # update this object with the values extracted from the payload
        if isinstance(data, str):
            _logger.debug(f'importing from filename {data}')
            data = yaml.safe_load(data)

        # this is REQUIRED to be separate to the return call
        new.inject_payload(data, limit)

        return new

    def inject_payload(self,
                       payload: dict,
                       limit: bool = True):
        """
        inject payload into the __dict__, effectively re-creating the object

        .. note ::
            use this function to unpack _within_ an object

        Args:
            payload (dict):
                __dict__ payload from the object that was packaged
            limit (bool):
                enable or disable safeties
        """
        if hasattr(self, '_logger'):
            self._logger.info(f'finalising unpacking of {type(self)}')

        # temporary attribute to check for a valid class
        self._unpack_validate = True

        if CLASS_STORAGE_KEY in payload:
            selfkey = get_class_storage(self)
            packkey = payload[CLASS_STORAGE_KEY]

            if selfkey != packkey:
                raise RuntimeError(f'attempting to unpack class '
                                   f'{packkey["name"]} as '
                                   f'{selfkey["name"]}')

            delattr(self, '_unpack_validate')

        unpacked = {k: self.unserialise(v, limit) for k, v in payload.items()}

        self.__dict__.update(unpacked)

    def serialise(self,
                  obj):
        """
        Recurse over any iterable objects, or call the pack() method of any
        `SendableMixin` objects, for serialisation

        Args:
            obj:
                object to be packaged

        Returns (yaml-serialisable object):
            yaml-friendly object
        """
        if issubclass(type(obj), SendableMixin):
            payload = obj.pack()
            return payload

        elif isinstance(obj, collections.abc.Mapping):
            # dict type
            try:
                payload = {self.serialise(k):
                               self.serialise(v) for k, v in obj.items()}
                return payload
            except TypeError as e:
                raise ValueError("Found an iterable dict-key. These cause "
                                 "circular issues with YAML")

        elif isinstance(obj, list):
            # list-type
            return [self.serialise(v) for v in obj]

        elif isinstance(obj, tuple):
            # tuple-type
            return [TUPLE_STORAGE_KEY] + [self.serialise(v) for v in obj]

        elif isinstance(obj, set):
            # set-type
            return [SET_STORAGE_KEY] + [self.serialise(v) for v in obj]

        return obj

    def unserialise(self,
                    obj,
                    limit: bool = True):
        """
        Undo a serialised packaging, by importing the called object and calling
        its unpack() method

        Args:
            obj:
                payload to be unpacked
            limit (bool):
                exit if the object to be unpacked is not part of the
                remotemanager package

        Returns:
            object before packaging
        """
        try:
            unpackable = obj.get(CLASS_STORAGE_KEY)
        except AttributeError:
            unpackable = False

        if unpackable:
            # print(f'unserialising unpackable object {obj}')
            # extract the class to import
            source = obj.pop(CLASS_STORAGE_KEY)

            # import the module
            modulename = source['mod']
            if limit and not modulename.startswith('remotemanager'):
                raise ValueError('module to import is not within the '
                                 'remotemanager package, exiting for safety')
            mod = importlib.import_module(modulename)

            # now get the actual class to import and unpack
            cls = getattr(mod, source['name'])

            return cls.unpack(obj)

        elif isinstance(obj, collections.abc.Mapping):
            # dict type
            return {self.unserialise(k):
                        self.unserialise(v) for k, v in obj.items()}

        # coming from the yaml file, output should _only_ be list
        elif isinstance(obj, list):
            try:
                if obj[0] == TUPLE_STORAGE_KEY:
                    return tuple(obj[1:])
                elif obj[0] == SET_STORAGE_KEY:
                    return set(obj[1:])
            except IndexError:
                return [self.unserialise(v) for v in obj]

        return obj


def get_class_storage(obj) -> dict:
    """
    Breaks down object into its module and classname.

    Args:
        obj:
            Python object to be broken down

    Returns (dict):
        module and classname dict
    """
    return {'mod': obj.__module__,
            'name': obj.__class__.__name__}


def get_mro_classnames(obj) -> list:
    return [subobj.__name__ for subobj in obj.__class__.__mro__]



"""
Handles file transfer via the `rsync` protocol
"""
import logging

from remotemanager.transport.transport import Transport
from remotemanager.utils.flags import Flags


class rsync(Transport):
    def __init__(self, *args, **kwargs):

        # flags can be exposed, to utilise their flexibility
        flags = kwargs.pop('flags', 'auv')
        self.flags = Flags(flags)

        super().__init__(*args, **kwargs)

        self._logger.info('created new rsync transport')

        self._cmd = 'rsync {flags} {primary}{files} {secondary}'

    @property
    def cmd(self):
        base = self._cmd.format(flags=self.flags.flags,
                                primary='{primary}',
                                files='{files}',
                                secondary='{secondary}')
        self._logger.debug(f'returning semi-formatted cmd: "{base}"')
        return base
